import json
import re
import unittest
import os

from poste_italiane_parser import get_parser


def list_tests():
    """
    Lists all test cases in the current module.
    """
    test_dir = os.path.dirname(__file__)
    return [
        os.path.join(test_dir, f)
        for f in os.listdir(test_dir)
        if f != "example.test.json" and f.endswith(".test.json")
    ]


class TestPIExtractor(unittest.TestCase):
    def test_all(self):
        tests = list_tests()

        if not tests:
            self.fail("No test cases found.")

        for test_file_path in tests:
            with open(test_file_path, "r") as f:
                with self.subTest(
                    msg=f"Testing file: {os.path.basename(test_file_path)}"
                ):
                    test_result = json.loads(f.read())

                    provider = get_parser(test_result["path"])

                    currency = provider.currency()
                    self.assertEqual(currency, test_result["currency"])

                    generated_at = provider.generated_at()
                    self.assertEqual(generated_at, test_result["generated_at"])

                    if "account_number" in test_result and (
                        test_result["account_number"] is not None
                        and test_result["account_number"] != ""
                    ):
                        account_number = provider.account_number()
                        self.assertEqual(account_number, test_result["account_number"])

                    period = provider.period()
                    self.assertEqual(
                        period["start_date"], test_result["period_start_date"]
                    )
                    self.assertEqual(period["end_date"], test_result["period_end_date"])

                    customer = provider.customer()
                    self.assertEqual(customer["name"], test_result["customer_name"])
                    if "street" in customer and (
                        customer["street"] is not None and customer["street"] != ""
                    ):
                        self.assertEqual(
                            customer["street"], test_result["customer_street"]
                        )
                    if "city" in customer and (
                        customer["city"] is not None and customer["city"] != ""
                    ):
                        self.assertEqual(customer["city"], test_result["customer_city"])

                    holder = provider.holder()
                    self.assertEqual(holder, test_result["holder"])

                    if "initial_balance" in test_result and (
                        test_result["initial_balance"] is not None
                        and test_result["initial_balance"] != ""
                    ):
                        initial_balance = provider.initial_balance()
                        self.assertEqual(
                            initial_balance, (test_result["initial_balance"])
                        )

                    if "final_balance" in test_result and (
                        test_result["final_balance"] is not None
                        and test_result["final_balance"] != ""
                    ):
                        final_balance = provider.final_balance()
                        self.assertEqual(final_balance, (test_result["final_balance"]))

                    if "card_number" in test_result and (
                        test_result["card_number"] is not None
                        and test_result["card_number"] != ""
                    ):
                        card_number = provider.card_number()
                        self.assertEqual(card_number, test_result["card_number"])

                    if "iban" in test_result and (
                        test_result["iban"] is not None and test_result["iban"] != ""
                    ):
                        iban = provider.iban()
                        self.assertEqual(iban, test_result["iban"])

                    # test transactions if are present
                    if "transactions" in test_result:
                        for transaction in test_result["transactions"]:
                            found = False
                            for t in provider.transactions():
                                for key in transaction:
                                    if key not in t or t[key] != transaction[key]:
                                        break
                                else:
                                    found = True
                                    break

                            self.assertTrue(
                                found,
                                f"Transaction {transaction} not found in provider transactions.",
                            )



from poste_italiane_parser.base_parser import BaseParser
from poste_italiane_parser.utils.pdf import parse_text_in_area


class ListaMovimentiParser(BaseParser):
    def __init__(self, path):
        super().__init__(
            name="LISTA_MOVIMENTI",
            path=path,
            parsing_config={
                "columns": {
                    "data": [80, 100],
                    "valuta": [130, 160],
                    "operazione": [170, 490],
                    "addebiti": [480, 510],
                    "accrediti": [540, 600],
                },
                "page_area": [0, 69, -1, 800],
                "first_page_area": [0, 238, -1, 800],
                "holder_area": [120, 154, 400, 168],
                "generated_at_area": [400, 77, 550, 90],
                "customer_area": [125, 152, 280, 158],
                "card_number_area": [120, 140, 220, 160],
                "iban_area": [240, 140, 380, 160],
            },
        )

    def parse(self):
        return super().parse()

    @staticmethod
    def check(path):
        """
        Checks if the PDF file at the given path is valid for this provider.
        """
        if not BaseParser.check(path):
            return False

        return parse_text_in_area(path, [40, 60, 200, 100]) == "SALDO E MOVIMENTI"



import re

from poste_italiane_parser.base_parser import BaseParser
from poste_italiane_parser.utils.pdf import parse_text_in_area


class RendicontoParser(BaseParser):
    def __init__(self, path):
        super().__init__(
            name="RENDICONTO",
            path=path,
            parsing_config={
                "columns": {
                    "data": [46, 100],
                    "valuta": [110, 160],
                    "addebiti": [220, 250],
                    "accrediti": [280, 330],
                    "operazione": [338, 580],
                },
                "page_area": [0, 130, -1, 726],
                "first_page_area": [0, 300, -1, 726],
                "holder_area": [365, 85, 500, 96],
                "generated_at_area": [400, 47, 460, 60],
                "customer_area": [290, 180, 510, 220],
                "card_number_area": [358, 74, 480, 85],
                "iban_area": [77, 134, 240, 145],
            },
        )

    def parse(self):
        return super().parse()

    @staticmethod
    def check(path):
        """
        Checks if the PDF file at the given path is valid for this provider.
        """
        if not BaseParser.check(path):
            return False

        area = [300, 70, 600, 80]
        account_regex = re.compile(r"\*{12}\d{4}")  # card number
        text = parse_text_in_area(path, area)

        return account_regex.search(text) if text else False



import datetime
import locale
import logging
import os
import re
from abc import ABC

from poste_italiane_parser.utils.pdf import parse_table, parse_text_in_area

logging.getLogger(__name__)

# Set the locale to Italian for date parsing
locale.setlocale(locale.LC_TIME, "it_IT.UTF-8")


class BaseParser(ABC):
    """
    This class serves as a base for all parsers, providing common functionality.
    """

    def __init__(self, name, path, parsing_config):
        if not self.check(path):
            raise ValueError(f"Invalid PDF file for parser: {path}")

        self.name = name

        self.parsing_config = parsing_config

        self.load(path)

    def load(self, path):
        self.path = path

        # get the rows from the PDF
        rows = parse_table(
            self.path,
            columns=self.parsing_config["columns"],
            page_area=self.parsing_config["page_area"],
            first_page_area=self.parsing_config["first_page_area"],
            required_columns=["data", "operazione"],
            multiple_page_rows=["operazione"],
            stop_fn=lambda value, values: value["operazione"] == "SALDO FINALE",
        )

        # map the rows to the desired format
        rows = [
            {
                "accounting_date": BaseParser._timestamp(row["data"]),
                "value_date": BaseParser._timestamp(row["valuta"]),
                "description": row["operazione"],
                "debits": BaseParser._float(row["addebiti"]),
                "credits": BaseParser._float(row["accrediti"]),
                "value": BaseParser._float(row["accrediti"])
                - BaseParser._float(row["addebiti"]),
            }
            for row in rows
        ]

        # filter by operation
        skip_operations = [
            "SALDO FINALE",
            "SALDO INIZIALE",
            "GIACENZA MEDIA NEL PERIODO",
            "TOTALE ENTRATE",
            "TOTALE USCITE",
        ]
        self.rows = [row for row in rows if row["description"] not in skip_operations]
        self.meta_rows = [row for row in rows if row["description"] in skip_operations]

        logging.debug(
            f"Parsed {len(self.rows)} rows and {len(self.meta_rows)} meta rows from {self.name}."
        )

        # validate the result of parsing
        # if not meta_rows we cannot validate the result
        # if self.meta_rows, we can sum all credits and check if "saldo finale" is equal to the sum of credits and debits
        if len(self.meta_rows) > 0:
            initial_balance = next(
                (
                    row["credits"]
                    for row in self.meta_rows
                    if row["description"] == "SALDO INIZIALE"
                ),
                None,
            )
            if not initial_balance:
                raise ValueError("Initial balance not found in the PDF.")

            final_balance = next(
                (
                    row["credits"]
                    for row in self.meta_rows
                    if row["description"] == "SALDO FINALE"
                ),
                None,
            )
            if not final_balance:
                raise ValueError("Final balance not found in the PDF.")

            calculated_total_income = sum(row["credits"] for row in self.rows)
            calculated_total_expense = sum(row["debits"] for row in self.rows)
            difference = calculated_total_income - calculated_total_expense
            expected_final_balance = round(initial_balance + difference, 4)

            if expected_final_balance != final_balance:
                raise ValueError(
                    f"Final balance mismatch: expected {expected_final_balance}, got {final_balance}"
                )

            logging.debug(
                f"Final balance matches: {expected_final_balance} == {final_balance}"
            )

            total_income = next(
                (
                    row["credits"]
                    for row in self.meta_rows
                    if row["description"] == "TOTALE ENTRATE"
                ),
                None,
            )
            total_expense = next(
                (
                    row["debits"]
                    for row in self.meta_rows
                    if row["description"] == "TOTALE USCITE"
                ),
                None,
            )

            if total_income is not None and total_expense is not None:
                logging.debug(
                    f"Total income: {total_income}, Total expense: {total_expense}, "
                    f"Calculated total income: {calculated_total_income}, "
                    f"Calculated total expense: {calculated_total_expense}"
                )

                if total_income != calculated_total_income:
                    raise ValueError(
                        f"Total income mismatch: expected {total_income}, got {calculated_total_income}"
                    )
                if total_expense != calculated_total_expense:
                    raise ValueError(
                        f"Total expense mismatch: expected {total_expense}, got {calculated_total_expense}"
                    )

    def holder(self):
        """
        Returns the holder of the account
        """
        return parse_text_in_area(self.path, self.parsing_config["holder_area"])

    def generated_at(self):
        """
        Returns the date when the document was generated.
        """
        date = parse_text_in_area(self.path, self.parsing_config["generated_at_area"])
        if not date:
            raise ValueError("Generated date not found in the PDF.")

        return BaseParser._timestamp(date)

    def customer(self):
        """
        Returns the customer information
        """
        customer = parse_text_in_area(self.path, self.parsing_config["customer_area"])

        customer = customer.split("\n")
        if len(customer) < 3:
            return {"name": customer[0].strip()}

        customer = {
            "name": customer[0].strip(),
            "street": customer[1].strip(),
            "city": customer[2].strip(),
        }

        return customer

    def currency(self):
        """
        Returns the currency of the account
        """
        if "currency_area" not in self.parsing_config:
            return "EURO"  # like in EstrattoContoParser

        curr = parse_text_in_area(self.path, self.parsing_config["currency_area"])
        if not curr:
            raise ValueError("Currency not found in the PDF.")

        return curr.strip().upper()

    def initial_balance(self):
        """
        Returns the initial balance
        """
        initial_balance = next(
            (
                row["credits"]
                for row in self.meta_rows
                if row["description"] == "SALDO INIZIALE"
            ),
            None,
        )

        return initial_balance

    def final_balance(self):
        """
        Returns the final balance
        """
        final_balance = next(
            (
                row["credits"]
                for row in self.meta_rows
                if row["description"] == "SALDO FINALE"
            ),
            None,
        )

        return final_balance

    def account_number(self):
        if "account_number_area" not in self.parsing_config:
            return None

        account = parse_text_in_area(
            self.path, self.parsing_config["account_number_area"]
        )
        if not account:
            raise ValueError("Account number not found in the PDF.")

        return account

    def card_number(self):
        """
        Returns the card number if available.
        """
        if "card_number_area" not in self.parsing_config:
            return None

        card_number = parse_text_in_area(
            self.path, self.parsing_config["card_number_area"]
        )
        if not card_number:
            return None

        # Find all characters that are digits (\d) or asterisks (\*)
        matches = re.findall(r"[\d\*]", card_number)
        return "".join(matches)

    def iban(self):
        """
        Returns the IBAN if available.
        """
        if "iban_area" not in self.parsing_config:
            return None

        iban = parse_text_in_area(self.path, self.parsing_config["iban_area"])
        return iban.strip().replace(" ", "") if iban else None

    def period(self):
        """
        Returns the period of the account
        """
        # check in the meta rows "SALDO INIZIALE" and "SALDO FINALE", if it exists, use the dates from these rows, otherwise use the first and last row of self.rows
        initial_balance_data = next(
            (
                row["accounting_date"]
                for row in self.meta_rows
                if row["description"] == "SALDO INIZIALE"
            ),
            None,
        )
        final_balance_data = next(
            (
                row["accounting_date"]
                for row in self.meta_rows
                if row["description"] == "SALDO FINALE"
            ),
            None,
        )

        if (not initial_balance_data or not final_balance_data) and not self.rows:
            return {
                "start_date": None,
                "end_date": None,
            }

        # if initial_balance_date start_date initial_balance_date else self.rows[0]["accounting_date"]
        start_date = (
            initial_balance_data
            if initial_balance_data
            else self.rows[0]["accounting_date"]
        )
        end_date = (
            final_balance_data
            if final_balance_data
            else self.rows[-1]["accounting_date"]
        )

        return {
            "start_date": start_date,
            "end_date": end_date,
        }

    def transactions(self):
        """
        Returns the parsed transactions from the PDF.
        """
        return self.rows

    def parse(self):
        """
        Parse the PDF file and return the parsed data.
        """
        return {
            "document_type": self.name,
            "currency": self.currency(),
            "account_number": self.account_number(),
            "card_number": self.card_number(),
            "iban": self.iban(),
            "customer": self.customer(),
            "holder": self.holder(),
            "generated_at": self.generated_at(),
            "initial_balance": self.initial_balance(),
            "final_balance": self.final_balance(),
            "period": self.period(),
            "transactions": self.transactions(),
        }

    @staticmethod
    def check(path):
        """
        Check if the given path is a valid PDF file for this parser.
        """
        return path.lower().endswith(".pdf") and os.path.isfile(path)

    @staticmethod
    def _float(value):
        return (
            float(value.replace("_", "").replace(".", "").replace(",", "."))
            if value
            else 0
        )

    @staticmethod
    def _timestamp(date):
        """
        Convert a date string in the format 'DD/MM/YY' or 'DD/MM/YYYY' or "DD MMMM YYYY" to a timestamp.
        """
        if not date:
            return None

        date = date.lower().strip()
        format = "%d/%m/%Y"

        if "ore" in date:  # listamovimenti
            format = "%d %B %Y ore %H:%M"
        else:
            date_parts = date.split("/")
            # if the date_parts are less than 2, there was an error in parsing

            date_parts[2] = (
                len(date_parts[2]) == 2 and f"20{date_parts[2]}" or date_parts[2]
            )
            date = "/".join(date_parts)

        try:
            date_obj = datetime.datetime.strptime(date, format).replace(
                tzinfo=datetime.timezone.utc
            )
            return date_obj.strftime("%Y-%m-%d %H:%M:%S")
        except ValueError:
            raise ValueError("Invalid date format.")



from poste_italiane_parser.estratto_conto import EstrattoContoParser
from poste_italiane_parser.lista_movimenti import ListaMovimentiParser
from poste_italiane_parser.rendiconto import RendicontoParser


def get_parser(path):
    """
    Find the appropriate parser for the given PDF file path.
    """
    if EstrattoContoParser.check(path):
        return EstrattoContoParser(path)
    elif RendicontoParser.check(path):
        return RendicontoParser(path)
    elif ListaMovimentiParser.check(path):
        return ListaMovimentiParser(path)
    else:
        return None


def PosteItalianeParser(path):
    """
    Parse the PDF file at the given path and return the parsed data.
    """
    provider = get_parser(path)
    if not provider:
        raise ValueError(f"No provider found for file: {path}")

    return provider.parse()



import re

from poste_italiane_parser.base_parser import BaseParser
from poste_italiane_parser.utils.pdf import parse_text_in_area


class EstrattoContoParser(BaseParser):
    def __init__(self, path):
        super().__init__(
            name="ESTRATTO_CONTO",
            path=path,
            parsing_config={
                "columns": {
                    "data": [70, 75],
                    "valuta": [120, 125],
                    "addebiti": [210, 230],
                    "accrediti": [300, 320],
                    "operazione": [360, 600],
                },
                "page_area": [0, 100, -1, 800],
                "first_page_area": [0, 300, -1, 800],
                "holder_area": [360, 82, 500, 94],
                "currency_area": [390, 68, 420, 82],
                "generated_at_area": [390, 40, 430, 50],
                "customer_area": [290, 180, 510, 220],
                "account_number_area": [150, 200, 220, 220],
                "iban_area": [36, 160, 240, 168],
            },
        )

    @staticmethod
    def check(path):
        """
        To determine if the given PDF is a Bancoposta Estratto Conto,
        we check if it is a valid PDF and if it contains an account number.
        """
        if not BaseParser.check(path):
            return False

        area = [150, 200, 220, 220]
        account_regex = re.compile(r"\d{1,14}")  # account number
        text = parse_text_in_area(path, area)

        return account_regex.search(text) if text else False



import re

import pymupdf


def parse_text_in_area(path, area):
    """
    Get the text from the first page of a PDF file within a specified area.
    """
    with pymupdf.open(path) as doc:
        if doc.page_count == 0:
            return None

        page = doc[0]
        area = get_area(page.rect.width, page.rect.height, area)

        tp = page.get_textpage(clip=area)
        text = tp.extractText()

        if not text:
            return None

        text = re.sub(r" {2,}", " ", text).strip()
        return text


def parse_table(
    path,
    columns=[],
    page_area=[0, 0, -1, -1],
    first_page_area=None,
    pages=None,
    required_columns=[],
    multiple_page_rows=[],
    stop_fn=None,
):
    """
    Parses a PDF file and extracts text values from specified columns within a given area.
    Args:
        path (str): The path to the PDF file.
        columns (dict): A dictionary mapping column names to their x-coordinate ranges.
        page_area (list): A list representing the area of the page to search for text values. Format: [x1, y1, x2, y2].
        first_page_area (list): A list representing the area of the first page to search for text values. Format: [x1, y1, x2, y2].
        pages (list or number): A list representing the range of pages to parse. Format: [start_page, end_page].
        required_columns (list): A list of column names that must be present in each extracted value.
        multiple_page_rows (list): A list of column names that, if present together, indicate a possible multiple-row value.
        stop_fn (function): A function that takes a current value and list of extracted values and returns True if the parsing should stop.
    Returns:
        list: A list of dictionaries, where each dictionary represents a row of extracted values. The keys of the dictionaries are the column names, and the values are the extracted text values.
    """
    values = []
    with pymupdf.open(path) as doc:
        pages = (
            [1, -1]
            if pages is None
            else ([pages, -1] if isinstance(pages, int) else pages)
        )
        pages[0] = pages[0] if pages[0] != -1 else doc.page_count
        pages[1] = pages[1] if pages[1] != -1 else doc.page_count
        if len(required_columns) == 0:
            required_columns = columns.keys()
        # Iterate through the pages of the PDF
        for page_num, page in enumerate(doc):
            page_num += 1
            if page_num < pages[0] or page_num > pages[1]:
                continue

            area = get_area(
                page.rect.width,
                page.rect.height,
                (first_page_area if page_num == 1 and first_page_area else page_area),
            )

            page_data = page.get_text("dict", clip=area)
            # Iterate through the blocks of text in the page
            for block in page_data["blocks"]:
                if "lines" not in block:
                    continue

                value = {}
                for line in block["lines"]:
                    for span in line["spans"]:
                        text = span["text"].strip()
                        bbox = span["bbox"]

                        x = bbox[2]
                        for key, (x1, x2) in columns.items():
                            if x >= x1 and x <= x2 and text != "":
                                if key not in value:
                                    value[key] = ""
                                value[key] += " " + text
                                value[key] = value[key].strip()

                is_multiple_page_rows = (
                    len(multiple_page_rows) > 0
                    and list(value.keys()) == multiple_page_rows
                )

                # validation
                if (
                    not all(k in value for k in required_columns)
                    and not is_multiple_page_rows
                ):
                    continue
                # Add "multiple_page_rows" values to the last row
                if is_multiple_page_rows:
                    if len(values) == 0:
                        continue
                    # Merge with the last value in values
                    last_index = len(values) - 1
                    last_value = values[last_index]
                    for k in multiple_page_rows:
                        last_value[k] += " " + value[k]
                        last_value[k] = last_value[k].strip()
                    values[last_index] = last_value
                else:
                    # add empty values for missing columns
                    for k in columns.keys():
                        if k not in value:
                            value[k] = ""
                        value[k] = value[k].strip()
                    values.append(value)
                    if stop_fn and stop_fn(value, values):
                        break
    return values


def get_area(page_width, page_height, area):
    area = list(area)
    area[0] = area[0] if area[0] != -1 else page_width
    area[1] = area[1] if area[1] != -1 else page_height
    area[2] = area[2] if area[2] != -1 else page_width
    area[3] = area[3] if area[3] != -1 else page_height

    return pymupdf.Rect(area)
